@@ -37,7 +37,7 @@
 
         Field 1: Constant beginning with "#".
 
-        Other fields are not defined and tuch a record is simply
+        Other fields are not defined and such a record is simply
         skipped during processing.
 
    1.2. Version record
@@ -94,7 +94,13 @@
 #include <dirent.h>
 #include <fcntl.h>
 #include <unistd.h>
+#ifndef HAVE_W32_SYSTEM
 #include <sys/utsname.h>
+#endif
+#ifdef MKDIR_TAKES_ONE_ARG
+#undef mkdir
+#define mkdir(a,b) mkdir(a)
+#endif
 
 #include "dirmngr.h"
 #include "validate.h"
@@ -555,9 +561,29 @@ open_dir (crl_cache_t *r_cache)
                   break;
                 }
             }
-          line = NULL; 
-          *entrytail = entry;
-          entrytail = &entry->next;
+
+          if (!entry->issuer_hash)
+            {
+              log_info (_("invalid line detected in `%s' line %u\n"),
+                        fname, lineno);
+              xfree (entry);
+              entry = NULL;
+            }
+          else if (find_entry (cache->entries, entry->issuer_hash))
+            {
+              /* Fixme: The duplicate checking used is not very
+                 effective for large numbers of issuers. */
+              log_info (_("duplicate entry detected in `%s' line %u\n"),
+                        fname, lineno);
+              xfree (entry);
+              entry = NULL;
+            }
+          else
+            {
+              line = NULL; 
+              *entrytail = entry;
+              entrytail = &entry->next;
+            }
         }
       else if (*line == '#')
         ;
@@ -712,12 +738,19 @@ update_dir (crl_cache_t cache)
   {
     char *tmpbuf, *p;
     const char *nodename;
+#ifndef HAVE_W32_SYSTEM
     struct utsname utsbuf;
+#endif
     
+#ifdef HAVE_W32_SYSTEM
+    nodename = "unknown";
+#else
     if (uname (&utsbuf))
       nodename = "unknown";
     else
       nodename = utsbuf.nodename;
+#endif
+
     tmpbuf = xmalloc (strlen (nodename) + 100);
     sprintf (tmpbuf, "DIR-tmp-%s-%u-%p.txt.tmp",
              nodename, (unsigned int)getpid (), tmpbuf);
@@ -742,13 +775,16 @@ update_dir (crl_cache_t cache)
       if (*line == 'c' || *line == 'i')
         {
           /* Extract the issuer hash field. */
-          char * endp = strchr (line+2, ':');
+          char *fieldp, *endp;
+
+          fieldp = strchr (line, ':');
+          endp = fieldp? strchr (++fieldp, ':') : NULL;
           if (endp)
             {
               /* There should be no percent within the issuer hash
                  field, thus we can compare it pretty easily. */
               *endp = 0;
-              e = find_entry ( cache->entries, line + 2);
+              e = find_entry ( cache->entries, fieldp);
               *endp = ':'; /* Restore orginal line. */
               if (e && e->deleted) 
                 {
@@ -765,6 +801,9 @@ update_dir (crl_cache_t cache)
                 { /* We ignore entries we don't have in our cache
                      because they may have been added in the meantime
                      by other instances of dirmngr. */
+                  fprintf (fpout, "# Next line added by "
+                           "another process; our pid is %lu\n", 
+                           (unsigned long)getpid ());
                   fputs (line, fpout);
                   putc ('\n', fpout);
                 }
@@ -796,6 +835,7 @@ update_dir (crl_cache_t cache)
             e->mark = 0;
           }
     }
+
   if (ferror (fp))
     {
       err = gpg_error_from_errno (errno);
@@ -1156,6 +1196,8 @@ cache_isvalid (ctrl_t ctrl, const char *issuer_hash,
   dirmngr_isotime_t current_time;
   size_t n;
 
+  (void)ctrl;
+
   entry = find_entry (cache->entries, issuer_hash);
   if (!entry)
     {
@@ -1516,6 +1558,8 @@ crl_parse_insert (ctrl_t ctrl, ksba_crl_t crl,
   gcry_md_hd_t md = NULL;
   int algo = 0;
   size_t n;
+  
+  (void)fname;
 
   *r_crlissuer = NULL;
   *thisupdate = *nextupdate = 0;
@@ -1546,9 +1590,16 @@ crl_parse_insert (ctrl_t ctrl, ksba_crl_t crl,
                 goto failure;
               }
 
-            if (opt.verbose)
+            if (opt.verbose || !*nextupdate)
               log_info (_("update times of this CRL: this=%s next=%s\n"), 
                         thisupdate, nextupdate);   
+            if (!*nextupdate)
+              {
+                log_info (_("nextUpdate not given; "
+                            "assuming a validity period of one day\n"));
+                copy_time (nextupdate, thisupdate);
+                add_isotime (nextupdate, 86400);
+              }
           }
           break;
       
@@ -1813,12 +1864,19 @@ crl_cache_insert (ctrl_t ctrl, const char *url, ksba_reader_t reader)
   {
     char *tmpfname, *p;
     const char *nodename;
+#ifndef HAVE_W32_SYSTEM
     struct utsname utsbuf;
-    
+#endif
+
+#ifdef HAVE_W32_SYSTEM
+    nodename = "unknown";
+#else
     if (uname (&utsbuf))
       nodename = "unknown";
     else
       nodename = utsbuf.nodename;
+#endif
+
     tmpfname = xmalloc (strlen (nodename) + 100);
     sprintf (tmpfname, "crl-tmp-%s-%u-%p.db.tmp",
              nodename, (unsigned int)getpid (), tmpfname);
