@@ -1,6 +1,6 @@
 /* crlfetch.c - LDAP access
  *      Copyright (C) 2002 Klarälvdalens Datakonsult AB
- *      Copyright (C) 2003, 2004, 2005 g10 Code GmbH
+ *      Copyright (C) 2003, 2004, 2005, 2006 g10 Code GmbH
  *
  * This file is part of DirMngr.
  *
@@ -16,19 +16,81 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
+ * 02110-1301, USA.
  */
 
 #include <config.h>
 
 #include <stdio.h>
 #include <errno.h>
+#include <pth.h>
 
 #include "crlfetch.h"
 #include "dirmngr.h"
 #include "misc.h"
 #include "http.h"
 
+#include "estream.h"
+
+/* We need to associate a reader object with file streams.  This table
+   is used for it. */
+struct file_reader_map_s
+{
+  ksba_reader_t reader;
+  estream_t fp;
+};
+#define MAX_FILE_READER 50
+static struct file_reader_map_s file_reader_map[MAX_FILE_READER];
+
+/* Associate FP with READER.  If the table is full wait until another
+   thread has removed an entry.  */
+static void
+register_file_reader (ksba_reader_t reader, estream_t fp)
+{
+  int i;
+  
+  for (;;)
+    {
+      for (i=0; i < MAX_FILE_READER; i++)
+        if (!file_reader_map[i].reader)
+          {
+            file_reader_map[i].reader = reader;
+            file_reader_map[i].fp = fp;
+            return;
+          }
+      log_info (_("reader to file mapping table full - waiting\n"));
+      pth_sleep (2); 
+    }
+}
+
+/* Scan the table for an entry matching READER, emove tha entry and
+   return the associated file pointer. */
+static estream_t
+get_file_reader (ksba_reader_t reader)
+{
+  estream_t fp = NULL;
+  int i;
+
+  for (i=0; i < MAX_FILE_READER; i++)
+    if (file_reader_map[i].reader == reader)
+      {
+        fp = file_reader_map[i].fp;
+        file_reader_map[i].reader = NULL;
+        file_reader_map[i].fp = NULL;
+        break;
+      }
+  return fp;
+}
+
+
+
+static int 
+my_es_read (void *opaque, char *buffer, size_t nbytes, size_t *nread)
+{
+  return es_read (opaque, buffer, nbytes, nread);
+}
+           
 
 /* Fetch CRL from URL and return the entire CRL using new ksba reader
    object in READER.  Note that this reader object should be closed
@@ -39,9 +101,11 @@ crl_fetch (ctrl_t ctrl, const char *url, ksba_reader_t *reader)
   gpg_error_t err;
   parsed_uri_t uri;
   char *free_this = NULL;
+  int redirects_left = 2; /* We allow for 2 redirect levels.  */
 
   *reader = NULL;
 
+ once_more:
   err = http_parse_uri (&uri, url);
   http_release_parsed_uri (uri);
   if (err && url && !strncmp (url, "https:", 6))
@@ -65,7 +129,7 @@ crl_fetch (ctrl_t ctrl, const char *url, ksba_reader_t *reader)
     }
   if (!err) /* Yes, our HTTP code groks that. */
     {
-      struct http_context_s hd;
+      http_t hd;
       
       if (opt.disable_http)
         {
@@ -74,40 +138,76 @@ crl_fetch (ctrl_t ctrl, const char *url, ksba_reader_t *reader)
           err = gpg_error (GPG_ERR_NOT_SUPPORTED);
         }
       else
-        err = http_open_document (&hd, url,
+        err = http_open_document (&hd, url, NULL,
                                   (opt.honor_http_proxy? HTTP_FLAG_TRY_PROXY:0)
-                                  |HTTP_FLAG_NO_SHUTDOWN,
-                                  opt.http_proxy);
+                                  |HTTP_FLAG_NO_SHUTDOWN
+                                  |HTTP_FLAG_NEED_HEADER, 
+                                  opt.http_proxy, NULL);
 
-      if (err)
-        log_error (_("error retrieving `%s': %s\n"), url, gpg_strerror (err));
-      else if (hd.status_code != 200)
+      switch ( err? 99999 : http_get_status_code (hd) )
         {
+        case 200:
+          {
+            estream_t fp = http_get_read_ptr (hd);
+            
+            err = ksba_reader_new (reader);
+            if (!err)
+	      err = ksba_reader_set_cb (*reader, &my_es_read, fp);
+            if (err)
+              {
+                log_error (_("error initializing reader object: %s\n"),
+                           gpg_strerror (err));
+                ksba_reader_release (*reader);
+                *reader = NULL;
+                http_close (hd, 0);
+              }
+            else
+              {
+                register_file_reader (*reader, fp);
+                http_close (hd, 1);
+              }
+          }
+          break;
+        
+        case 301: /* Redirection (perm.). */
+        case 302: /* Redirection (temp.). */
+          {
+            const char *s = http_get_header (hd, "Location");
+
+            log_info (_("URL `%s' redirected to `%s' (%u)\n"),
+                      url, s?s:"[none]", http_get_status_code (hd));
+            if (s && *s && redirects_left-- )
+              {
+                xfree (free_this); url = NULL;
+                free_this = xtrystrdup (s);
+                if (!free_this)
+                  err = gpg_error_from_errno (errno);
+                else
+                  {
+                    url = free_this;
+                    http_close (hd, 0);
+                    /* Note, that our implementation of redirection
+                       actually handles a redirect to LDAP.  */
+                    goto once_more;
+                  }
+              }
+            else
+              err = gpg_error (GPG_ERR_NO_DATA);
+            log_error (_("too many redirections\n")); /* Or no "Location". */
+            http_close (hd, 0);
+          }
+          break;
+  
+        case 99999: /* Made up status code foer error reporting.  */
+          log_error (_("error retrieving `%s': %s\n"),
+                     url, gpg_strerror (err));
+          break;
+
+        default:
           log_error (_("error retrieving `%s': http status %u\n"),
-                     url, hd.status_code);
+                     url, http_get_status_code (hd));
           err = gpg_error (GPG_ERR_NO_DATA);
-          http_close (&hd, 0);
-        }
-      else
-        {
-          err = ksba_reader_new (reader);
-          if (!err)
-            err = ksba_reader_set_file (*reader, hd.fp_read);
-          if (err)
-            {
-              log_error (_("error initializing reader object: %s\n"),
-                         gpg_strerror (err));
-              ksba_reader_release (*reader);
-              *reader = NULL;
-              http_close (&hd, 0);
-            }
-          else
-            {
-              /* FIXME: Register the fp somewhere so that we are able
-                 to release the ksba_reader and close the underlying
-                 stream. See also below crl_close_reader(). */
-              http_close (&hd, 1);
-            }
+          http_close (hd, 0);
         }
     }
   else /* Let the LDAP code try other schemes. */
@@ -121,8 +221,8 @@ crl_fetch (ctrl_t ctrl, const char *url, ksba_reader_t *reader)
       else
         err = url_fetch_ldap (ctrl, url, NULL, 0, reader);
     }
-  if (free_this)
-    xfree (free_this);
+
+  xfree (free_this);
   return err;
 }
 
@@ -231,15 +331,17 @@ end_cert_fetch (cert_fetch_context_t context)
 void
 crl_close_reader (ksba_reader_t reader)
 {
+  estream_t fp;
+
   if (!reader)
     return;
 
   /* Check whether this is a HTTP one. */
-
-  /* FIXME: TODO */ 
-
-  /* If not a ldap wrapper context might be assocaited.  */
-  ldap_wrapper_release_context (reader);
+  fp = get_file_reader (reader);
+  if (fp) /* This is an HTTP context. */
+    es_fclose (fp);
+  else /* This is an ldap wrapper context. */
+    ldap_wrapper_release_context (reader);
 
   /* Now get rid of the reader object. */
   ksba_reader_release (reader);
