@@ -1,6 +1,6 @@
 /* dirmngr.c - LDAP access
  *	Copyright (C) 2002 Klarälvdalens Datakonsult AB
- *      Copyright (C) 2003, 2004, 2006 g10 Code GmbH
+ *      Copyright (C) 2003, 2004, 2006, 2007 g10 Code GmbH
  *
  * This file is part of DirMngr.
  *
@@ -16,7 +16,8 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
+ * 02110-1301, USA.
  */
 
 #include <config.h>
@@ -30,9 +31,11 @@
 #include <assert.h>
 #include <time.h>
 #include <fcntl.h>
+#ifndef HAVE_W32_SYSTEM
 #include <sys/socket.h>
-#include <sys/stat.h>
 #include <sys/un.h>
+#endif
+#include <sys/stat.h>
 #include <unistd.h>
 #include <signal.h>
 #include <pth.h>
@@ -47,6 +50,12 @@
 #include "crlcache.h"
 #include "crlfetch.h"
 #include "misc.h"
+#include "i18n.h"
+
+#ifdef HAVE_W32_SYSTEM
+#include "../jnlib/w32-afunix.h"
+#define sleep _sleep
+#endif
 
 enum cmd_and_opt_values 
 { aNull = 0,
@@ -58,12 +67,14 @@ enum cmd_and_opt_values
   
   aServer,
   aDaemon,
+  aService,
   aListCRLs,
   aLoadCRL,
   aFetchCRL,
   aShutdown,
   aFlush,
   aGPGConfList,
+  aGPGConfTest,
 
   oOptions,
   oDebug,
@@ -91,6 +102,7 @@ enum cmd_and_opt_values
   oOCSPResponder,
   oOCSPSigner,
   oOCSPMaxClockSkew,
+  oOCSPMaxPeriod,
   oOCSPCurrentPeriod,
   oMaxReplies,
   oFakedSystemTime,
@@ -109,12 +121,16 @@ static ARGPARSE_OPTS opts[] = {
 
   { aServer,   "server",    256, N_("run in server mode (foreground)") },
   { aDaemon,   "daemon",    256, N_("run in daemon mode (background)") },
+#if HAVE_W32_SYSTEM
+  { aService,  "service",   256, N_("run as windows service (background)") },
+#endif
   { aListCRLs, "list-crls", 256, N_("list the contents of the CRL cache")},
   { aLoadCRL,  "load-crl",  256, N_("|FILE|load CRL from FILE into cache")},
   { aFetchCRL, "fetch-crl", 256, N_("|URL|fetch a CRL from URL")},
   { aShutdown, "shutdown",  256, N_("shutdown the dirmngr")},
   { aFlush,    "flush",     256, N_("flush the cache")},
   { aGPGConfList, "gpgconf-list", 256, "@" },
+  { aGPGConfTest, "gpgconf-test", 256, "@" },
 
   { 301, NULL, 0, N_("@\nOptions:\n ") },
 
@@ -159,6 +175,7 @@ static ARGPARSE_OPTS opts[] = {
   { oOCSPResponder, "ocsp-responder", 2, N_("|URL|use OCSP responder at URL")},
   { oOCSPSigner, "ocsp-signer", 2, N_("|FPR|OCSP response signed by FPR")}, 
   { oOCSPMaxClockSkew, "ocsp-max-clock-skew", 1, "@" },
+  { oOCSPMaxPeriod, "ocsp-max-period", 1, "@" },
   { oOCSPCurrentPeriod, "ocsp-current-period", 1, "@" },
 
   { oMaxReplies, "max-replies", 1,
@@ -185,7 +202,6 @@ static ARGPARSE_OPTS opts[] = {
 
 #define DEFAULT_MAX_REPLIES 10
 #define DEFAULT_LDAP_TIMEOUT 100 /* arbitrary large timeout */
-#define DEFAULT_SOCKET_NAME "/var/run/dirmngr/socket"
 
 /* For the cleanup handler we need to keep track of the socket's name. */
 static const char *socket_name;
@@ -194,19 +210,28 @@ static int cleanup_socket;
 /* Keep track of the current log file so that we can avoid updating
    the log file after a SIGHUP if it didn't changed. Malloced. */
 static char *current_logfile;
-/* Name of a config file, which will be reread on a HUP if it is not NULL. */
-static char *config_filename;
 /* Helper to implement --debug-level. */
 static const char *debug_level;
 /* Flag indicating that a shutdown has been requested.  */
-static int shutdown_pending;
+static volatile int shutdown_pending;
 /* Counter for the active connections.  */
 static int active_connections;
 
+/* The timer tick used for housekeeping stuff.  For Windows we use a
+   longer period as the SetWaitableTimer seems to signal earlier than
+   the 2 seconds.  */
+#ifdef HAVE_W32_SYSTEM
+#define TIMERTICK_INTERVAL    (4)
+#else
+#define TIMERTICK_INTERVAL    (2)    /* Seconds.  */
+#endif
+
+
 /* Prototypes. */
 static void cleanup (void);
 static ldap_server_t parse_ldapserver_file (const char* filename);
 static void free_ldapservers_list (ldap_server_t servers);
+static fingerprint_list_t parse_ocsp_signer (const char *string);
 static void handle_connections (int listen_fd);
 
 /* Pth wrapper function definitions. */
@@ -216,28 +241,29 @@ GCRY_THREAD_OPTION_PTH_IMPL;
 static const char *
 my_strusage( int level )
 {
-    const char *p;
-    switch( level ) {
-      case 11: p = "dirmngr";
-	break;
-      case 13: p = VERSION; break;
-      case 14: p = "Copyright (C) 2006 g10 Code GmbH"; break;
-      case 17: p = PRINTABLE_OS_NAME; break;
-      case 19: p =
-	    _("Please report bugs to <gpa-dev@gnupg.org>.\n");
-	break;
-      case 1:
-      case 40:	p =
-	    _("Usage: dirmngr [options] (-h for help)");
-	break;
-      case 41:	p =
-	    _("Syntax: dirmngr [options] [command [args]]\n"
-	      "LDAP and OCSP access for GnuPG\n");
-	break;
-
-      default:	p = NULL;
+  const char *p;
+  switch ( level ) 
+    {
+    case 11: p = "dirmngr";
+      break;
+    case 13: p = VERSION; break;
+    case 14: p = "Copyright (C) " COPYRIGHT_YEAR_NAME; break;
+    case 17: p = PRINTABLE_OS_NAME; break;
+      /* TRANSLATORS: @EMAIL@ will get replaced by the actual bug
+         reporting address.  This is so that we can change the
+         reporting address without breaking the translations.  */
+    case 19: p = _("Please report bugs to <@EMAIL@>.\n"); break;
+    case 49: p = PACKAGE_BUGREPORT; break;
+    case 1:
+    case 40: p = _("Usage: dirmngr [options] (-h for help)");
+      break;
+    case 41: p = _("Syntax: dirmngr [options] [command [args]]\n"
+                   "LDAP and OCSP access for GnuPG\n");
+      break;
+      
+    default: p = NULL;
     }
-    return p;
+  return p;
 }
 
 
@@ -245,7 +271,9 @@ my_strusage( int level )
 static void
 my_gcry_logger (void *dummy, int level, const char *fmt, va_list arg_ptr)
 {
-  /* translate the log levels */
+  (void)dummy;
+
+  /* Translate the log levels. */
   switch (level)
     {
     case GCRY_LOG_CONT: level = JNLIB_LOG_CONT; break;
@@ -270,6 +298,8 @@ my_ksba_hash_buffer (void *arg, const char *oid,
                      const void *buffer, size_t length, size_t resultsize,
                      unsigned char *result, size_t *resultlen)
 {
+  (void)arg;
+
   if (oid && strcmp (oid, "1.3.14.3.2.26")) 
     return gpg_error (GPG_ERR_NOT_SUPPORTED); 
   if (resultsize < 20)
@@ -404,10 +434,16 @@ parse_rereadable_options (ARGPARSE_ARGS *pargs, int reread)
       opt.ignore_ocsp_service_url = 0;
       opt.allow_ocsp = 0;
       opt.ocsp_responder = NULL;
-      opt.ocsp_signer = NULL; 
       opt.ocsp_max_clock_skew = 10 * 60;      /* 10 minutes.  */
+      opt.ocsp_max_period = 90 * 86400;       /* 90 days.  */
       opt.ocsp_current_period = 3 * 60 * 60;  /* 3 hours. */
       opt.max_replies = DEFAULT_MAX_REPLIES;
+      while (opt.ocsp_signer)
+        {
+          fingerprint_list_t tmp = opt.ocsp_signer->next;
+          xfree (opt.ocsp_signer);
+          opt.ocsp_signer = tmp;
+        }
       return 1;
     }
 
@@ -450,8 +486,11 @@ parse_rereadable_options (ARGPARSE_ARGS *pargs, int reread)
 
     case oAllowOCSP: opt.allow_ocsp = 1; break;
     case oOCSPResponder: opt.ocsp_responder = pargs->r.ret_str; break;
-    case oOCSPSigner:    opt.ocsp_signer = pargs->r.ret_str; break;
+    case oOCSPSigner: 
+      opt.ocsp_signer = parse_ocsp_signer (pargs->r.ret_str);
+      break;
     case oOCSPMaxClockSkew: opt.ocsp_max_clock_skew = pargs->r.ret_int; break;
+    case oOCSPMaxPeriod: opt.ocsp_max_period = pargs->r.ret_int; break;
     case oOCSPCurrentPeriod: opt.ocsp_current_period = pargs->r.ret_int; break;
 
     case oMaxReplies: opt.max_replies = pargs->r.ret_int; break;
@@ -463,9 +502,49 @@ parse_rereadable_options (ARGPARSE_ARGS *pargs, int reread)
   return 1; /* Handled. */
 }
 
-int
-main (int argc, char **argv )
+
+#ifdef HAVE_W32_SYSTEM
+/* The global status of our service.  */
+SERVICE_STATUS_HANDLE service_handle;
+SERVICE_STATUS service_status;
+
+DWORD WINAPI
+w32_service_control (DWORD control, DWORD event_type, LPVOID event_data,
+		     LPVOID context)
 {
+  /* event_type and event_data are not used here.  */
+  switch (control)
+    {
+    case SERVICE_CONTROL_SHUTDOWN:
+      /* For shutdown we will try to force termination.  */
+      service_status.dwCurrentState = SERVICE_STOP_PENDING;
+      SetServiceStatus (service_handle, &service_status); 
+      shutdown_pending = 3;
+      break;
+
+    case SERVICE_CONTROL_STOP:
+      service_status.dwCurrentState = SERVICE_STOP_PENDING;
+      SetServiceStatus (service_handle, &service_status); 
+      shutdown_pending = 1;
+      break;
+
+    default:
+      break;
+    }
+  return 0;
+}
+#endif /*HAVE_W32_SYSTEM*/
+
+
+#ifdef HAVE_W32_SYSTEM
+#define main real_main
+#endif
+int
+main (int argc, char **argv)
+{
+#ifdef HAVE_W32_SYSTEM
+#undef main
+#endif
   enum cmd_and_opt_values cmd = 0;
   ARGPARSE_ARGS pargs;
   int orig_argc;
@@ -486,6 +565,29 @@ main (int argc, char **argv )
   int rc;
   int homedir_seen = 0;
 
+#ifdef HAVE_W32_SYSTEM
+  /* The option will be set by main() below if we should run as a
+     system daemon.  */
+  if (opt.system_service)
+    {
+      service_handle
+	= RegisterServiceCtrlHandlerEx ("DirMngr",
+					&w32_service_control, NULL /*FIXME*/);
+      if (service_handle == 0)
+	log_error ("failed to register service control handler: ec=%d",
+		   (int) GetLastError ());
+      service_status.dwServiceType = SERVICE_WIN32_OWN_PROCESS;
+      service_status.dwCurrentState = SERVICE_START_PENDING;
+      service_status.dwControlsAccepted
+	= SERVICE_ACCEPT_STOP | SERVICE_ACCEPT_SHUTDOWN;
+      service_status.dwWin32ExitCode = NO_ERROR;
+      service_status.dwServiceSpecificExitCode = NO_ERROR;
+      service_status.dwCheckPoint = 0;
+      service_status.dwWaitHint = 10000; /* 10 seconds timeout.  */
+      SetServiceStatus (service_handle, &service_status); 
+    }
+#endif /*HAVE_W32_SYSTEM*/
+
   set_strusage (my_strusage);
   log_set_prefix ("dirmngr", 1|4); 
   
@@ -527,6 +629,7 @@ main (int argc, char **argv )
   assuan_set_malloc_hooks (gcry_malloc, gcry_realloc, gcry_free);
   assuan_set_assuan_log_stream (log_get_stream ());
   assuan_set_assuan_log_prefix (log_get_prefix (NULL));
+  assuan_set_assuan_err_source (GPG_ERR_SOURCE_DEFAULT);
 
   /* Setup I18N. */
   i18n_init();
@@ -536,15 +639,7 @@ main (int argc, char **argv )
   if (shell && strlen (shell) >= 3 && !strcmp (shell+strlen (shell)-3, "csh") )
     csh_style = 1;
   
-  opt.homedir = getenv("GNUPGHOME");
-  if (!opt.homedir || !*opt.homedir)
-    {
-#ifdef HAVE_DRIVE_LETTERS
-      opt.homedir = "c:/gnupg";
-#else
-      opt.homedir = "~/.gnupg";
-#endif
-    }
+  opt.homedir = default_homedir ();
 
   /* Reset rereadable options to default values. */
   parse_rereadable_options (NULL, 0); 
@@ -554,7 +649,7 @@ main (int argc, char **argv )
   opt.ldaptimeout = DEFAULT_LDAP_TIMEOUT;
 
   /* Other defaults.  */
-  socket_name = DEFAULT_SOCKET_NAME;
+  socket_name = default_socket_name ();
 
 
   /* Check whether we have a config file given on the commandline */
@@ -582,6 +677,12 @@ main (int argc, char **argv )
         }
       else if (pargs.r_opt == aDaemon)
         opt.system_daemon = 1;
+      else if (pargs.r_opt == aService)
+        {
+	  /* Redundant.  The main function takes care of it.  */
+	  opt.system_service = 1;
+	  opt.system_daemon = 1;
+	}
     }
 
   /* If --daemon has been given on the command line but not --homedir,
@@ -589,9 +690,9 @@ main (int argc, char **argv )
      this also overrides the GNUPGHOME environment variable.  */
   if (opt.system_daemon && !homedir_seen)
     {
-      opt.homedir = DIRMNGR_SYSCONFDIR;
-      opt.homedir_data = DIRMNGR_DATADIR;
-      opt.homedir_cache = DIRMNGR_CACHEDIR;
+      opt.homedir = dirmngr_sysconfdir ();
+      opt.homedir_data = dirmngr_datadir ();
+      opt.homedir_cache = dirmngr_cachedir ();
     }
 
   if (default_config)
@@ -637,12 +738,14 @@ main (int argc, char **argv )
         {
         case aServer: 
         case aDaemon:
+        case aService:
         case aShutdown: 
         case aFlush: 
 	case aListCRLs: 
 	case aLoadCRL: 
         case aFetchCRL:
 	case aGPGConfList:
+	case aGPGConfTest:
           cmd = pargs.r_opt;
           break;
 
@@ -694,7 +797,7 @@ main (int argc, char **argv )
       fclose( configfp );
       configfp = NULL;
       /* Keep a copy of the name so that it can be read on SIGHUP. */
-      config_filename = configname;
+      opt.config_filename = configname;
       configname = NULL;
       goto next_pass;
     }
@@ -743,15 +846,18 @@ main (int argc, char **argv )
   else
       opt.ldapservers = parse_ldapserver_file (ldapfile);
 
+#ifndef HAVE_W32_SYSTEM
   /* We need to ignore the PIPE signal because the we might log to a
      socket and that code handles EPIPE properly.  The ldap wrapper
      also requires us to ignore this silly signal. Assuan would set
      this signal to ignore anyway.*/
   signal (SIGPIPE, SIG_IGN);
-
+#endif
 
   /* Ready.  Now to our duties. */
-  if (!cmd)
+  if (!cmd && opt.system_service)
+    cmd = aDaemon;
+  else if (!cmd)
     cmd = aServer;
   rc = 0;
 
@@ -800,18 +906,24 @@ main (int argc, char **argv )
           current_logfile = xstrdup (logfile);
         }
 
-      if (strchr (socket_name, ':') )
+#ifndef HAVE_W32_SYSTEM
+      if (strchr (socket_name, ':'))
         {
           log_error (_("colons are not allowed in the socket name\n"));
           dirmngr_exit (1);
         }
+#endif
       if (strlen (socket_name)+1 >= sizeof serv_addr.sun_path ) 
         {
           log_error (_("name of socket too long\n"));
           dirmngr_exit (1);
         }
-   
+    
+#ifdef HAVE_W32_SYSTEM
+      fd = _w32_sock_new (AF_UNIX, SOCK_STREAM, 0);
+#else
       fd = socket (AF_UNIX, SOCK_STREAM, 0);
+#endif
       if (fd == -1)
         {
           log_error (_("can't create socket: %s\n"), strerror(errno) );
@@ -825,12 +937,21 @@ main (int argc, char **argv )
       len = (offsetof (struct sockaddr_un, sun_path)
              + strlen(serv_addr.sun_path) + 1);
 
+#ifdef HAVE_W32_SYSTEM
+      rc = _w32_sock_bind (fd, (struct sockaddr*) &serv_addr, len);
+      if (rc == -1 && errno == WSAEADDRINUSE)
+	{
+	  remove (socket_name);
+	  rc = _w32_sock_bind (fd, (struct sockaddr*) &serv_addr, len);
+	}
+#else
       rc = bind (fd, (struct sockaddr*)&serv_addr, len);
       if (rc == -1 && errno == EADDRINUSE)
         {
           remove (socket_name);
           rc = bind (fd, (struct sockaddr*)&serv_addr, len);
         }
+#endif
       if (rc == -1)
         {
           log_error (_("error binding socket to `%s': %s\n"),
@@ -851,6 +972,11 @@ main (int argc, char **argv )
         log_info (_("listening on socket `%s'\n"), socket_name );
 
       fflush (NULL);
+
+#ifdef HAVE_W32_SYSTEM
+      pid = getpid ();
+      printf ("set DIRMNGR_INFO=%s;%lu;1\n", socket_name, (ulong) pid);
+#else
       pid = pth_fork ();
       if (pid == (pid_t)-1) 
         {
@@ -925,13 +1051,28 @@ main (int argc, char **argv )
               dirmngr_exit (1);
             }
         }
+#endif
 
       launch_reaper_thread ();
       cert_cache_init ();
       crl_cache_init ();
+#ifdef HAVE_W32_SYSTEM
+      if (opt.system_service)
+	{
+	  service_status.dwCurrentState = SERVICE_RUNNING;
+	  SetServiceStatus (service_handle, &service_status);
+	}
+#endif
       handle_connections (fd);
       close (fd);
       shutdown_reaper ();
+#ifdef HAVE_W32_SYSTEM
+      if (opt.system_service)
+	{
+	  service_status.dwCurrentState = SERVICE_STOPPED;
+	  SetServiceStatus (service_handle, &service_status);
+	}
+#endif
     }
   else if (cmd == aListCRLs)
     {
@@ -994,6 +1135,8 @@ main (int argc, char **argv )
         wrong_args ("--flush");
       rc = crl_cache_flush();
     }
+  else if (cmd == aGPGConfTest)
+    dirmngr_exit (0);
   else if (cmd == aGPGConfList)
     {
       char *filename;
@@ -1015,11 +1158,12 @@ main (int argc, char **argv )
 
       /* First the configuration file.  This is not an option, but it
 	 is vital information for GPG Conf.  */
-      if (!config_filename)
-        config_filename = make_filename (opt.homedir, "dirmngr.conf", NULL );
+      if (!opt.config_filename)
+        opt.config_filename = make_filename (opt.homedir,
+                                             "dirmngr.conf", NULL );
 
       printf ("gpgconf-dirmngr.conf:%lu:\"%s\n",
-              GC_OPT_FLAG_DEFAULT, config_filename);
+              GC_OPT_FLAG_DEFAULT, opt.config_filename);
 
       printf ("verbose:%lu:\n", GC_OPT_FLAG_NONE);
       printf ("quiet:%lu:\n", GC_OPT_FLAG_NONE);
@@ -1069,6 +1213,40 @@ main (int argc, char **argv )
 }
 
 
+#ifdef HAVE_W32_SYSTEM
+int
+main (int argc, char *argv[])
+{
+  int i;
+
+  /* Find out if we run in daemon mode or on the command line.  */
+  for (i = 1; i < argc; i++)
+    if (!strcmp (argv[i], "--service"))
+      {
+	opt.system_service = 1;
+	opt.system_daemon = 1;
+	break;
+      }
+
+  if (!opt.system_service)
+    return real_main (argc, argv);
+  else
+    {
+      SERVICE_TABLE_ENTRY DispatchTable [] =
+	{
+	  /* Ignore warning.  */
+	  { "DirMngr", &real_main },
+	  { NULL, NULL }
+	};
+
+      if (!StartServiceCtrlDispatcher (DispatchTable))
+        return 1;
+      return 0;
+    }
+}
+#endif
+
+
 static void
 cleanup (void)
 {
@@ -1098,6 +1276,8 @@ dirmngr_exit (int rc)
 void
 dirmngr_init_default_ctrl (ctrl_t ctrl)
 {
+  (void)ctrl;
+
   /* Nothing for now. */
 }
 
@@ -1253,6 +1433,129 @@ free_ldapservers_list (ldap_server_t servers)
 }
 
 
+static fingerprint_list_t
+parse_ocsp_signer (const char *string)
+{
+  gpg_error_t err;
+  char *fname;
+  FILE *fp;
+  char line[256];
+  char *p;
+  fingerprint_list_t list, *list_tail, item;
+  unsigned int lnr = 0;
+  int c, i, j;
+  int errflag = 0;
+
+
+  /* Check whether this is not a filename and treat it as a direct
+     fingerprint specification.  */
+  if (!strpbrk (string, "/.~\\"))
+    {
+      item = xcalloc (1, sizeof *item);
+      for (i=j=0; (string[i] == ':' || hexdigitp (string+i)) && j < 40; i++)
+        if ( string[i] != ':' )
+          item->hexfpr[j++] = string[i] >= 'a'? (string[i] & 0xdf): string[i];
+      item->hexfpr[j] = 0;
+      if (j != 40 || !(spacep (string+i) || !string[i]))
+        {
+          log_error (_("%s:%u: invalid fingerprint detected\n"), 
+                     "--ocsp-signer", 0);
+          xfree (item);
+          return NULL;
+        }
+      return item;
+    }
+  
+  /* Well, it is a filename.  */
+  if (*string == '/' || (*string == '~' && string[1] == '/'))
+    fname = make_filename (string, NULL);
+  else
+    {
+      if (string[0] == '.' && string[1] == '/' )
+        string += 2;
+      fname = make_filename (opt.homedir, string, NULL);
+    }
+
+  fp = fopen (fname, "r");
+  if (!fp)
+    {
+      err = gpg_error_from_syserror ();
+      log_error (_("can't open `%s': %s\n"), fname, gpg_strerror (err));
+      xfree (fname);
+      return NULL;
+    }
+
+  list = NULL;
+  list_tail = &list;
+  for (;;)
+    {
+      if (!fgets (line, DIM(line)-1, fp) )
+        {
+          if (!feof (fp))
+            {
+              err = gpg_error_from_syserror ();
+              log_error (_("%s:%u: read error: %s\n"),
+                         fname, lnr, gpg_strerror (err));
+              errflag = 1;
+            }
+          fclose (fp);
+          if (errflag)
+            {
+              while (list)
+                {
+                  fingerprint_list_t tmp = list->next;
+                  xfree (list);
+                  list = tmp;
+                }
+            }
+          xfree (fname);
+          return list; /* Ready.  */
+        }
+
+      lnr++;
+      if (!*line || line[strlen(line)-1] != '\n')
+        {
+          /* Eat until end of line. */
+          while ( (c=getc (fp)) != EOF && c != '\n')
+            ;
+          err = gpg_error (*line? GPG_ERR_LINE_TOO_LONG
+                           /* */: GPG_ERR_INCOMPLETE_LINE);
+          log_error (_("%s:%u: read error: %s\n"), 
+                     fname, lnr, gpg_strerror (err));
+          errflag = 1;
+          continue;
+        }
+
+      /* Allow for empty lines and spaces */
+      for (p=line; spacep (p); p++)
+        ;
+      if (!*p || *p == '\n' || *p == '#')
+        continue;
+
+      item = xcalloc (1, sizeof *item);
+      *list_tail = item;
+      list_tail = &item->next;
+
+      for (i=j=0; (p[i] == ':' || hexdigitp (p+i)) && j < 40; i++)
+        if ( p[i] != ':' )
+          item->hexfpr[j++] = p[i] >= 'a'? (p[i] & 0xdf): p[i];
+      item->hexfpr[j] = 0;
+      if (j != 40 || !(spacep (p+i) || p[i] == '\n'))
+        {
+          log_error (_("%s:%u: invalid fingerprint detected\n"), fname, lnr);
+          errflag = 1;
+        }
+      i++;
+      while (spacep (p+i))
+        i++;
+      if (p[i] && p[i] != '\n')
+        log_info (_("%s:%u: garbage at end of line ignored\n"), fname, lnr);
+    }
+  /*NOTREACHED*/
+}
+
+
+
 
 /*
    Stuff used in daemon mode.  
@@ -1277,14 +1580,14 @@ reread_configuration (void)
   unsigned int configlineno = 0;
   int dummy;
 
-  if (!config_filename)
+  if (!opt.config_filename)
     return; /* No config file. */
 
-  fp = fopen (config_filename, "r");
+  fp = fopen (opt.config_filename, "r");
   if (!fp)
     {
       log_error (_("option file `%s': %s\n"),
-                 config_filename, strerror(errno) );
+                 opt.config_filename, strerror(errno) );
       return;
     }
 
@@ -1294,7 +1597,7 @@ reread_configuration (void)
   dummy = 0;
   pargs.argc = &dummy;
   pargs.flags = 1;  /* do not remove the args */
-  while (optfile_parse (fp, config_filename, &configlineno, &pargs, opts) )
+  while (optfile_parse (fp, opt.config_filename, &configlineno, &pargs, opts) )
     {
       if (pargs.r_opt < -1)
         pargs.err = 1; /* Print a warning. */
@@ -1314,6 +1617,7 @@ handle_signal (int signo)
 {
   switch (signo)
     {
+#ifndef HAVE_W32_SYSTEM
     case SIGHUP:
       log_info (_("SIGHUP received - "
                   "re-reading configuration and flushing caches\n"));
@@ -1354,13 +1658,34 @@ handle_signal (int signo)
       cleanup ();
       dirmngr_exit (0);
       break;
-
+#endif
     default:
       log_info (_("signal %d received - no action defined\n"), signo);
     }
 }
 
 
+/* This is the worker for the ticker.  It is called every few seconds
+   and may only do fast operations. */
+static void
+handle_tick (void)
+{
+  /* Nothing real to do right now.  Actually we need the timeout only
+     for W32 where we don't use signals and need a way for the loop to
+     check for the shutdown flag. */
+#ifdef HAVE_W32_SYSTEM
+  if (shutdown_pending)
+    log_info (_("SIGTERM received - shutting down ...\n"));
+  if (shutdown_pending > 2)
+    {
+      log_info (_("shutdown forced\n"));
+      log_info ("%s %s stopped\n", strusage(11), strusage(13) );
+      cleanup ();
+      dirmngr_exit (0);
+    }
+#endif /*HAVE_W32_SYSTEM*/
+}
+
 
 /* Helper to call a connection's main fucntion. */
 static void *
@@ -1387,8 +1712,8 @@ static void
 handle_connections (int listen_fd)
 {
   pth_attr_t tattr;
-  pth_event_t ev;
-  sigset_t sigs;
+  pth_event_t ev, time_ev;
+  sigset_t sigs, oldsigs;
   int signo;
   struct sockaddr_un paddr;
   socklen_t plen = sizeof( paddr );
@@ -1399,6 +1724,7 @@ handle_connections (int listen_fd)
   pth_attr_set (tattr, PTH_ATTR_STACK_SIZE, 1024*1024);
   pth_attr_set (tattr, PTH_ATTR_NAME, "dirmngr");
 
+#ifndef HAVE_W32_SYSTEM /* FIXME */
   sigemptyset (&sigs );
   sigaddset (&sigs, SIGHUP);
   sigaddset (&sigs, SIGUSR1);
@@ -1406,6 +1732,11 @@ handle_connections (int listen_fd)
   sigaddset (&sigs, SIGINT);
   sigaddset (&sigs, SIGTERM);
   ev = pth_event (PTH_EVENT_SIGS, &sigs, &signo);
+#else
+  sigs = 0;
+  ev = pth_event (PTH_EVENT_SIGS, &sigs, &signo);
+#endif
+  time_ev = NULL;
 
   for (;;)
     {
@@ -1423,12 +1754,29 @@ handle_connections (int listen_fd)
           continue;
 	}
 
+      if (!time_ev)
+        time_ev = pth_event (PTH_EVENT_TIME, 
+                             pth_timeout (TIMERTICK_INTERVAL, 0));
+
+      if (time_ev)
+        pth_event_concat (ev, time_ev, NULL);
       fd = pth_accept_ev (listen_fd, (struct sockaddr *)&paddr, &plen, ev);
+      if (time_ev)
+        pth_event_isolate (time_ev);
+
       if (fd == -1)
         {
-          if (pth_event_occurred (ev))
+          if (pth_event_occurred (ev)
+              || (time_ev && pth_event_occurred (time_ev)) )
             {
-              handle_signal (signo);
+              if (pth_event_occurred (ev))
+                handle_signal (signo);
+              if (time_ev && pth_event_occurred (time_ev))
+                {
+                  pth_event_free (time_ev, PTH_FREE_ALL);
+                  time_ev = NULL;
+                  handle_tick ();
+                }
               continue;
 	    }
           log_error (_("accept failed: %s - waiting 1s\n"), strerror (errno));
@@ -1436,15 +1784,39 @@ handle_connections (int listen_fd)
           continue;
 	}
 
+      if (pth_event_occurred (ev))
+        {
+          handle_signal (signo);
+        }
+
+      if (time_ev && pth_event_occurred (time_ev))
+        {
+          pth_event_free (time_ev, PTH_FREE_ALL);
+          time_ev = NULL;
+          handle_tick ();
+        }
+
+
+      /* We now might create a new thread and because we don't want
+         any signals (as we are handling them here) to be delivered to
+         a new thread we need to block those signals. */
+      pth_sigmask (SIG_BLOCK, &sigs, &oldsigs);
+
+      /* Create thread to handle this connection.  */
       if (!pth_spawn (tattr, start_connection_thread, (void*)fd))
         {
           log_error (_("error spawning connection handler: %s\n"),
                      strerror (errno) );
           close (fd);
 	}
+
+      /* Restore the signal mask. */
+      pth_sigmask (SIG_SETMASK, &oldsigs, NULL);
     }
   
   pth_event_free (ev, PTH_FREE_ALL);
+  if (time_ev)
+    pth_event_free (time_ev, PTH_FREE_ALL);
   cleanup ();
   log_info ("%s %s stopped\n", strusage(11), strusage(13));
 }
