@@ -51,6 +51,15 @@ struct ldap_server_s {
 typedef struct ldap_server_s *ldap_server_t;
 
 
+/* A list of fingerprints.  */
+struct fingerprint_list_s;
+typedef struct fingerprint_list_s *fingerprint_list_t;
+struct fingerprint_list_s
+{
+  fingerprint_list_t next;
+  char hexfpr[20+20+1];
+};
+
 
 /* A large struct name "opt" to keep global flags. */
 struct 
@@ -64,11 +73,15 @@ struct
   const char *homedir_data; /* Ditto for data files (/usr/share/dirmngr).  */
   const char *homedir_cache; /* Ditto for cache files (/var/cache/dirmngr).  */
 
+  char *config_filename;     /* Name of a config file, which will be
+                                reread on a HUP if it is not NULL. */
+
   char *ldap_wrapper_program; /* Override value for the LDAP wrapper
                                  program.  */
   char *http_wrapper_program; /* Override value for the HTTP wrapper
                                  program.  */
 
+  int system_service;   /* We are running as W32 service (implies daemon).  */
   int system_daemon;    /* We are running in system daemon mode.  */
   int running_detached; /* We are running in detached mode.  */
 
@@ -93,12 +106,15 @@ struct
   ldap_server_t ldapservers;
   int add_new_ldapservers;
 
-  const char *ocsp_responder; /* Standard OCSP responder's URL. */
-  const char *ocsp_signer;    /* The fingerprint of the standard OCSP
-                                 responder signer's certificate. */
-  unsigned int ocsp_max_clock_skew;  /* Allowed seconds of clocks skew. */
-  unsigned int ocsp_current_period;  /* Seconds a response is
-                                        considered current. */
+  const char *ocsp_responder;     /* Standard OCSP responder's URL. */
+  fingerprint_list_t ocsp_signer; /* The list of fingerprints with allowed
+                                     standard OCSP signer certificates.  */
+  
+  unsigned int ocsp_max_clock_skew; /* Allowed seconds of clocks skew. */
+  unsigned int ocsp_max_period;     /* Seconds a response is at maximum
+                                       considered valid after thisUpdate. */
+  unsigned int ocsp_current_period; /* Seconds a response is considered 
+                                       current after nextUpdate. */
 } opt;
 
 
