@@ -1,6 +1,6 @@
 /* ldap.c - LDAP access
  *      Copyright (C) 2002 Klarälvdalens Datakonsult AB
- *      Copyright (C) 2003, 2004, 2005 g10 Code GmbH
+ *      Copyright (C) 2003, 2004, 2005, 2007 g10 Code GmbH
  *
  * This file is part of DirMngr.
  *
@@ -27,16 +27,25 @@
 #include <errno.h>
 #include <unistd.h>
 #include <fcntl.h>
-#include <sys/wait.h>
 #include <time.h>
+#ifdef HAVE_W32_SYSTEM
+#include <winsock2.h>
+#include <winldap.h>
+#else
 #include <ldap.h> /* fixme: remove it */
+#endif
 #include <pth.h>
 #include <ksba.h> 
 
+#include "exechelp.h"
 #include "crlfetch.h"
 #include "dirmngr.h"
 #include "misc.h"
 
+#if HAVE_W32_SYSTEM
+#define setenv(a,b,c) SetEnvironmentVariable ((a),(b))
+#endif
+
 /* In case sysconf does not return a value we need to have a limit. */
 #ifdef _POSIX_OPEN_MAX
 #define MAX_OPEN_FDS _POSIX_OPEN_MAX
@@ -52,6 +61,7 @@
                             "$-_.+!*'(),"
 #define USERCERTIFICATE "userCertificate"
 #define CACERTIFICATE   "caCertificate"
+#define X509CACERT      "x509caCert"
 #define USERSMIMECERTIFICATE "userSMIMECertificate"
 
 
@@ -159,7 +169,7 @@ static void
 destroy_wrapper (struct wrapper_context_s *ctx) 
 {
   if (ctx->pid != (pid_t)(-1))
-    kill (ctx->pid, SIGTERM);
+    dirmngr_kill_process (ctx->pid);
   ksba_reader_release (ctx->reader);
   if (ctx->fd)
     close (ctx->fd);
@@ -242,6 +252,8 @@ ldap_wrapper_thread (void *dummy)
   char line[256];
   time_t current_time;
 
+  (void)dummy;
+
   for (;;)
     {
       FD_ZERO (&read_fds);
@@ -309,35 +321,30 @@ ldap_wrapper_thread (void *dummy)
           /* Check whether the process is still running.  */
           if (ctx->pid != (pid_t)(-1))
             {
-              int i, status;
+              gpg_error_t err;
+	      int status;
               
-              while ( (i=waitpid (ctx->pid, &status, WNOHANG)) == -1
-                      && errno == EINTR)
-                ;
-              if (i == -1)
-                log_error (_("waiting for ldap wrapper %d failed: %s\n"),
-                           (int)ctx->pid, strerror (errno));
-              else if (i)
+	      err = dirmngr_wait_process (ctx->pid, 0, &status);
+              if (!err)
                 {
-                  if (!WIFEXITED (status))
-                    log_info (_("ldap wrapper %d ready: terminated\n"),
-                              (int)ctx->pid);
-                  else if (WEXITSTATUS (status) == 10 )
-                    log_info (_("ldap wrapper %d ready: timeout\n"),
-                              (int)ctx->pid);
-                  else 
-                    log_info (_("ldap wrapper %d ready: exit status %d\n"),
-                              (int)ctx->pid, WEXITSTATUS (status));
+		  log_info (status == 10 ?
+                            _("ldap wrapper %d ready: timeout\n") :
+                            _("ldap wrapper %d ready"), (int)ctx->pid);
                   ctx->ready = 1;
                   ctx->pid = (pid_t)(-1);
                 }
+              else if (gpg_err_code (err) != GPG_ERR_TIMEOUT)
+                {
+                  log_error (_("waiting for ldap wrapper %d failed: %s\n"),
+                             (int)ctx->pid, gpg_strerror (err));
+                }
             }
 
           /* Check whether we should terminate the process. */
           if (ctx->pid != (pid_t)(-1)
               && ctx->stamp != (time_t)(-1) && ctx->stamp < current_time)
             {
-              if (!kill (ctx->pid, SIGTERM))
+              if (!dirmngr_kill_process (ctx->pid))
                 ctx->stamp = (time_t)(-1);
               log_info (_("ldap wrapper %d stalled - killing\n"),
                         (int)ctx->pid);
@@ -423,7 +430,7 @@ ldap_wrapper_connection_cleanup (ctrl_t ctrl)
         ctx->ctrl->refcount--;
         ctx->ctrl = NULL;
         if (ctx->pid != (pid_t)(-1))
-          kill (ctx->pid, SIGTERM);
+          dirmngr_kill_process (ctx->pid);
         if (ctx->fd_error)
           log_info (_("reading from ldap wrapper %d failed: %s\n"),
                     ctx->printable_pid, gpg_strerror (ctx->fd_error));
@@ -513,9 +520,8 @@ reader_callback (void *cb_value, char *buffer, size_t count,  size_t *nread)
 }
 
 /* Fork and exec the LDAP wrapper and returns a new libksba reader
-   object at READER.  ARGV is a NULL terminated list or argumenst for
-   the wrapper; however the function adds the program's name as the
-   first arg.  The function returns 0 on success or an error code.
+   object at READER.  ARGV is a NULL terminated list or arguments for
+   the wrapper.  The function returns 0 on success or an error code.
 
    We can't use LDAP directly for these reasons:
 
@@ -529,31 +535,35 @@ reader_callback (void *cb_value, char *buffer, size_t count,  size_t *nread)
    3. There is no easy way for timeouts. In particular the timeout
       value does not work for DNS lookups (well, this is usual) and it
       seems not to work while loading a large attribute like a
-      CRL. Having a separate process allows us to either tell than
+      CRL. Having a separate process allows us to either tell the
       process to commit suicide or have our own housekepping function
       kill it after some time.  The latter also allows proper
-      cancellation of a query at any point.
+      cancellation of a query at any point of time.
       
    4. Given that we are going out to the network and usually get back
-      a long response, the frok/exec overhead is acceptable.
+      a long response, the fork/exec overhead is acceptable.
 
    Special hack to avoid passing a password through the command line
-   which is gloabbally visible: If the first element of ARGV is
-   "--pass" it will be removed and instead the environment variable
+   which is globally visible: If the first element of ARGV is "--pass"
+   it will be removed and instead the environment variable
    DIRMNGR_LDAP_PASS will be set to the next value of ARGV.  On modern
-   OSes the environment is not visible to other other user.  For those
-   old systems where it can't be avoided, we don't want to go into the
+   OSes the environment is not visible to other users.  For those old
+   systems where it can't be avoided, we don't want to go into the
    hassle of passing the password via stdin; it's just too complicated
-   and an LDAP password used for public directory lookups should be
-   that confidential.
-  */
+   and an LDAP password used for public directory lookups should not
+   be that confidential.  */
 static gpg_error_t
 ldap_wrapper (ctrl_t ctrl, ksba_reader_t *reader, const char *argv[])
 {
   gpg_error_t err;
   pid_t pid;
-  int rp[2], rp2[2];
   struct wrapper_context_s *ctx;
+  int i;
+  int j;
+  char **arg_list;
+  char *pgmname;
+  int fd_out;
+  int fd_err;
 
   /* It would be too simple to connect stderr just to our logging
      stream.  The problem is that if we are running multi-threaded
@@ -566,127 +576,49 @@ ldap_wrapper (ctrl_t ctrl, ksba_reader_t *reader, const char *argv[])
 
   *reader = NULL;
 
-  if (pipe (rp) == -1)
-    {
-      err = gpg_error_from_errno (errno);
-      log_error (_("error creating a pipe: %s\n"), strerror (errno));
-      return err;
-    }
-  if (pipe (rp2) == -1)
-    {
-      err = gpg_error_from_errno (errno);
-      log_error (_("error creating a pipe: %s\n"), strerror (errno));
-      close (rp[0]);
-      close (rp[1]);
-      return err;
-    }
-      
-  pid = pth_fork ();
-  if (pid == (pid_t)(-1))
-    {
-      err = gpg_error_from_errno (errno);
-      log_error (_("error forking process: %s\n"), strerror (errno));
-      close (rp[0]);
-      close (rp[1]);
-      close (rp2[0]);
-      close (rp2[1]);
-      return err;
-    }
+  /* Files: We need to prepare stdin and stdout.  We get stderr from
+     the function.  */
+  if (!opt.ldap_wrapper_program || !*opt.ldap_wrapper_program)
+    pgmname = (char *) get_dirmngr_ldap_path ();
+  else
+    pgmname = opt.ldap_wrapper_program;
 
-  if (!pid)
-    { /* Child. */
-      char *pgmname;
-      char **arg_list;
-      int n, i, j;
-      int fd;
+  /* Create command line argument array.  */
+  for (i = 0; argv[i]; i++)
+    ;
+  arg_list = xcalloc (i + 2, sizeof *arg_list);
+  for (i = j = 0; argv[i]; i++, j++)
+    if (!i && argv[i + 1] && !strcmp (*argv, "--pass"))
+      {
+	arg_list[j] = "--env-pass";
+	setenv ("DIRMNGR_LDAP_PASS", argv[1], 1);
+	i++;
+      }
+    else
+      arg_list[j] = (char*) argv[i];
 
-      if (!opt.ldap_wrapper_program || !*opt.ldap_wrapper_program)
-        pgmname = DIRMNGR_LIBEXECDIR "/dirmngr_ldap";
-      else
-        pgmname = opt.ldap_wrapper_program;
-
-      /* Create command line argument array.  */
-      for (i=0; argv[i]; i++)
-        ;
-      arg_list = xcalloc (i+3, sizeof *arg_list);
-      arg_list[0] = strrchr (pgmname, '/');
-      if (arg_list[0])
-        arg_list[0]++;
-      else
-        arg_list[0] = pgmname;
-      for (i=0,j=1; argv[i]; i++, j++)
-        if (!i && argv[i+1] && !strcmp (*argv, "--pass"))
-          {
-            arg_list[j] = "--env-pass";
-            setenv ("DIRMNGR_LDAP_PASS", argv[1], 1 );
-            i++;
-          }
-        else
-          arg_list[j] = (char*)argv[i];
-
-      /* Connect stdin to /dev/null.  */
-      fd = open ("/dev/null", O_RDONLY);
-      if (fd == -1)
-        {
-          log_error (_("can't open `%s': %s\n"), "/dev/null",strerror (errno));
-          _exit (4);
-        }
-      if (fd != STDIN_FILENO && dup2 (fd, STDIN_FILENO) == -1)
-        {
-          log_error (_("dup2 failed in child: %s\n"), strerror (errno));
-          _exit (4);
-        }
-
-      /* Connect stdout to the first pipe.  */
-      if (rp[1] != STDOUT_FILENO && dup2 (rp[1], STDOUT_FILENO) == -1)
-        {
-          log_error (_("dup2 failed in child: %s\n"), strerror (errno));
-          _exit (4);
-        }
-
-      /* Connect stderr to the second pipe.  */
-      if (rp2[1] != STDERR_FILENO && dup2 (rp2[1], STDERR_FILENO) == -1)
-        {
-          log_error (_("dup2 failed in child: %s\n"), strerror (errno));
-          _exit (4);
-        }
-
-  
-      /* Close all files which will not be duped. */
-      n = sysconf (_SC_OPEN_MAX);
-      if (n < 0)
-        n = MAX_OPEN_FDS;
-      for (i=0; i < n; i++)
-        {
-          if ( i == STDIN_FILENO || i == STDOUT_FILENO || i == STDERR_FILENO)
-            continue;
-          close(i);
-        }
-      errno = 0;
-
-      execv (pgmname, arg_list);
-      log_error (_("error running `%s': %s\n"), pgmname, strerror (errno));
-      _exit (31);
-    }
-
-  /* Parent. */
-  close (rp[1]);
-  close (rp2[1]);
-  
   ctx = xtrycalloc (1, sizeof *ctx);
   if (!ctx)
     {
       err = gpg_error_from_errno (errno);
       log_error (_("error allocating memory: %s\n"), strerror (errno));
-      kill (pid, SIGTERM);
-      close (rp[0]);
-      close (rp2[0]);
+      xfree (arg_list);
       return err;
     }
+
+  err = dirmngr_spawn_process (pgmname, arg_list,
+			       &fd_out, &fd_err, &pid);
+  xfree (arg_list);
+  if (err)
+    {
+      xfree (ctx);
+      return err;
+    }
+
   ctx->pid = pid;
-  ctx->printable_pid = (int)pid;
-  ctx->fd = rp[0];
-  ctx->log_fd = rp2[0];
+  ctx->printable_pid = (int) pid;
+  ctx->fd = fd_out;
+  ctx->log_fd = fd_err;
   ctx->ctrl = ctrl;
   ctrl->refcount++;
   ctx->stamp = time (NULL);
@@ -1055,7 +987,8 @@ make_url (char **url, const char *dn, const char *filter)
   char *u_dn, *u_filter;
   char const attrs[] = (USERCERTIFICATE ","
 /*                         USERSMIMECERTIFICATE "," */
-                        CACERTIFICATE);
+                        CACERTIFICATE ","
+                        X509CACERT );
 
   *url = NULL;
 
@@ -1397,6 +1330,13 @@ fetch_next_cert_ldap (cert_fetch_context_t context,
                                CACERTIFICATE);
                   okay = 1;
                 }
+              else if (!ascii_strcasecmp (p, X509CACERT))
+                {
+                  if (DBG_LOOKUP)
+                    log_debug ("fetch_next_cert_ldap: got attribute `%s'\n",
+                               CACERTIFICATE);
+                  okay = 1;
+                }
 /*               else if (!ascii_strcasecmp (p, USERSMIMECERTIFICATE)) */
 /*                 { */
 /*                   if (DBG_LOOKUP) */
