@@ -45,38 +45,17 @@
    something reasonable. */
 #define MAX_CERT_LENGTH (8*1024)
 
-#define PARM_ERROR(t) assuan_set_error (ctx, ASSUAN_Parameter_Error, (t))
+#define PARM_ERROR(t) assuan_set_error (ctx, \
+                                        gpg_error (GPG_ERR_ASS_PARAMETER), (t))
 
 
 /* Data used to associate an Assuan context with local server data */
-struct server_local_s {
-  ASSUAN_CONTEXT assuan_ctx;
+struct server_local_s 
+{
+  assuan_context_t assuan_ctx;
 };
 
 
-/* Map GPG_ERR_xx error codes to Assuan status codes */
-static int
-map_to_assuan_status (int rc)
-{
-  gpg_err_code_t   ec = gpg_err_code (rc);
-  gpg_err_source_t es = gpg_err_source (rc);
-
-  if (!rc)
-    return 0;
-  if (!es)
-    {
-      es = GPG_ERR_SOURCE_USER_4; /*  This should not happen, but we
-                                      need to make sure to pass a new
-                                      Assuan error code along. */
-      log_debug ("map_to_assuan_status called with no error source\n");
-    }
-
-  if (ec == -1)
-    ec = GPG_ERR_NO_DATA;  /* That used to be ASSUAN_No_Data_Available. */
-
-  return gpg_err_make (es, ec);
-}
-
 
 
 /* Copy the % and + escaped string S into the buffer D and replace the
@@ -106,6 +85,53 @@ strcpy_escaped_plus (char *d, const unsigned char *s)
 }
 
 
+/* Check whether the option NAME appears in LINE */
+static int
+has_option (const char *line, const char *name)
+{
+  const char *s;
+  int n = strlen (name);
+
+  s = strstr (line, name);
+  return (s && (s == line || spacep (s-1)) && (!s[n] || spacep (s+n)));
+}
+
+
+/* Same as has_option but does only test for the name of the option
+   and ignores an argument, i.e. with NAME being "--hash" it would
+   return a pointer for "--hash" as well as for "--hash=foo".  If
+   thhere is no such option NULL is returned.  The pointer returned
+   points right behind the option name, this may be an equal sign, Nul
+   or a space.  */
+/* static const char * */
+/* has_option_name (const char *line, const char *name) */
+/* { */
+/*   const char *s; */
+/*   int n = strlen (name); */
+
+/*   s = strstr (line, name); */
+/*   return (s && (s == line || spacep (s-1)) */
+/*           && (!s[n] || spacep (s+n) || s[n] == '=')) ? (s+n) : NULL; */
+/* } */
+
+
+/* Skip over options.  It is assumed that leading spaces have been
+   removed (this is the case for lines passed to a handler from
+   assuan).  Blanks after the options are also removed. */
+static char *
+skip_options (char *line)
+{
+  while ( *line == '-' && line[1] == '-' )
+    {
+      while (*line && !spacep (line))
+        line++;
+      while (spacep (line))
+        line++;
+    }
+  return line;
+}
+
+
 /* Common code for get_cert_local and get_issuer_cert_local. */
 static ksba_cert_t 
 do_get_cert_local (ctrl_t ctrl, const char *name, const char *command)
@@ -130,7 +156,7 @@ do_get_cert_local (ctrl_t ctrl, const char *name, const char *command)
   if (rc)
     {
       log_error (_("assuan_inquire(%s) failed: %s\n"),
-                 command, assuan_strerror (rc));
+                 command, gpg_strerror (rc));
       return NULL;
     }
   
@@ -242,7 +268,7 @@ get_cert_local_ski (ctrl_t ctrl, const char *name, ksba_sexp_t keyid)
   if (rc)
     {
       log_error (_("assuan_inquire(%s) failed: %s\n"), "SENDCERT_SKI",
-                 assuan_strerror (rc));
+                 gpg_strerror (rc));
       return NULL;
     }
   
@@ -277,14 +303,13 @@ inquire_cert_and_load_crl (assuan_context_t ctx)
 {
   ctrl_t ctrl = assuan_get_pointer (ctx);
   gpg_error_t err;
-  assuan_error_t ae;
   unsigned char *value = NULL;
   size_t valuelen; 
   ksba_cert_t cert = NULL;
 
-  ae = assuan_inquire( ctx, "SENDCERT", &value, &valuelen, 0);
-  if (ae)
-    return map_assuan_err (ae);
+  err = assuan_inquire( ctx, "SENDCERT", &value, &valuelen, 0);
+  if (err)
+    return err;
 
 /*   { */
 /*     FILE *fp = fopen ("foo.der", "r"); */
@@ -315,7 +340,7 @@ inquire_cert_and_load_crl (assuan_context_t ctx)
 
 /* Handle OPTION commands. */
 static int
-option_handler (ASSUAN_CONTEXT ctx, const char *key, const char *value)
+option_handler (assuan_context_t ctx, const char *key, const char *value)
 {
   ctrl_t ctrl = assuan_get_pointer (ctx);
 
@@ -325,13 +350,14 @@ option_handler (ASSUAN_CONTEXT ctx, const char *key, const char *value)
       ctrl->force_crl_refresh = i;
     }
   else
-    return ASSUAN_Invalid_Option;
+    return gpg_error (GPG_ERR_UNKNOWN_OPTION);
 
   return 0;
 }
 
 
-/* IS_VALID <certificate_id>|<certificate_fpr>
+/* ISVALID [--only-ocsp] [--force-default-responder]
+            <certificate_id>|<certificate_fpr>
   
    This command checks whether the certificate identified by the
    certificate_id is valid.  This is done by consulting CRLs or
@@ -339,12 +365,19 @@ option_handler (ASSUAN_CONTEXT ctx, const char *key, const char *value)
    are from gpg-error.h.  The command may callback using the inquire
    function.  See the manual for details.
  
-   The certificate_id is a hex encoded string consisting of two parts,
+   The CERTIFICATE_ID is a hex encoded string consisting of two parts,
    delimited by a single dot.  The first part is the SHA-1 hash of the
    issuer name and the second part the serial number.
 
    Alternatively the certificate's fingerprint may be given in which
    case an OCSP request is done before consulting the CRL.
+
+   If the option --only-ocsp is given, no fallback to a CRL check will
+   be used.
+
+   If the option --force-default-responder is given, only the default
+   OCSP responder will be used and any other methods of obtaining an
+   OCSP responder URL won't be used.
  */
 
 static int
@@ -355,7 +388,13 @@ cmd_isvalid (assuan_context_t ctx, char *line)
   gpg_error_t err;
   int did_inquire = 0;
   int ocsp_mode = 0;
+  int only_ocsp;
+  int force_default_responder;
   
+  only_ocsp = has_option (line, "--only-ocsp");
+  force_default_responder = has_option (line, "--force-default-responder");
+  line = skip_options (line);
+
   issuerhash = xstrdup (line); /* We need to work on a copy of the
                                   line because that same Assuan
                                   context may be used for an inquiry.
@@ -389,13 +428,15 @@ cmd_isvalid (assuan_context_t ctx, char *line)
       if (!opt.allow_ocsp)
         err = gpg_error (GPG_ERR_NOT_SUPPORTED);
       else
-        err = ocsp_isvalid (ctrl, NULL, NULL);
-      /* Fixme: If we got no ocsp response we should fall back to CRL
-         mode.  Thus we need to clear OCSP_MODE, get the issuerhash
-         and the serialno from the current certificate and jump to
-         again. */
+        err = ocsp_isvalid (ctrl, NULL, NULL, force_default_responder);
+      /* Fixme: If we got no ocsp response and --only-ocsp is not used
+         we should fall back to CRL mode.  Thus we need to clear
+         OCSP_MODE, get the issuerhash and the serialno from the
+         current certificate and jump to again. */
     }
-  else
+  else if (only_ocsp)
+    err = gpg_error (GPG_ERR_NO_CRL_KNOWN);
+  else 
     {
       switch (crl_cache_isvalid (ctrl,
                                  issuerhash, serialno,
@@ -427,7 +468,7 @@ cmd_isvalid (assuan_context_t ctx, char *line)
   if (err)
     log_error (_("command %s failed: %s\n"), "ISVALID", gpg_strerror (err));
   xfree (issuerhash);
-  return map_to_assuan_status (err);
+  return err;
 }
 
 
@@ -503,16 +544,14 @@ cmd_checkcrl (assuan_context_t ctx, char *line)
     {
       /* We do not have this certificate yet or the fingerprint has
          not been given.  Inquire it from the client.  */
-      assuan_error_t ae;
       unsigned char *value = NULL;
       size_t valuelen; 
       
-      ae = assuan_inquire (ctrl->server_local->assuan_ctx, "TARGETCERT",
+      err = assuan_inquire (ctrl->server_local->assuan_ctx, "TARGETCERT",
                            &value, &valuelen, MAX_CERT_LENGTH);
-      if (ae)
+      if (err)
         {
-          log_error (_("assuan_inquire failed: %s\n"), assuan_strerror (ae));
-          err = map_assuan_err (ae);
+          log_error (_("assuan_inquire failed: %s\n"), gpg_strerror (err));
           goto leave;
         }
   
@@ -543,17 +582,17 @@ cmd_checkcrl (assuan_context_t ctx, char *line)
   if (err)
     log_error (_("command %s failed: %s\n"), "CHECKCRL", gpg_strerror (err));
   ksba_cert_release (cert);
-  return map_to_assuan_status (err);
+  return err;
 }
 
 
-/* CHECKOCSP [<fingerprint>]
+/* CHECKOCSP [--force-default-responder] [<fingerprint>]
 
    Check whether the certificate with FINGERPRINT (SHA-1 hash of the
    entire X.509 certificate blob) is valid or not by asking an OCSP
    responder responsible for this certificate.  The optional
    fingerprint may be used for a quick check in case an OCSP check has
-   been down for this certificate recently (we always cache OCSP
+   been done for this certificate recently (we always cache OCSP
    responses for a couple of minutes). If the fingerprint has not been
    given or there is no cached result, the function inquires the
    certificate using the
@@ -567,6 +606,10 @@ cmd_checkcrl (assuan_context_t ctx, char *line)
    its own mechanism which includes a local certificate store as well
    as a list of trusted root certifciates.
 
+   If the option --force-default-responder is given, only the default
+   OCSP responder will be used and any other methods of obtaining an
+   OCSP responder URL won't be used.
+
    The return value is the usual gpg-error code or 0 for ducesss;
    i.e. the certificate validity has been confirmed by a valid CRL.
 */
@@ -577,6 +620,10 @@ cmd_checkocsp (assuan_context_t ctx, char *line)
   gpg_error_t err;
   unsigned char fprbuffer[20], *fpr;
   ksba_cert_t cert;
+  int force_default_responder;
+  
+  force_default_responder = has_option (line, "--force-default-responder");
+  line = skip_options (line);
 
   fpr = get_fingerprint_from_line (line, fprbuffer);
   cert = fpr? get_cert_byfpr (fpr) : NULL;
@@ -585,16 +632,14 @@ cmd_checkocsp (assuan_context_t ctx, char *line)
     {
       /* We do not have this certificate yet or the fingerprint has
          not been given.  Inquire it from the client.  */
-      assuan_error_t ae;
       unsigned char *value = NULL;
       size_t valuelen; 
       
-      ae = assuan_inquire (ctrl->server_local->assuan_ctx, "TARGETCERT",
+      err = assuan_inquire (ctrl->server_local->assuan_ctx, "TARGETCERT",
                            &value, &valuelen, MAX_CERT_LENGTH);
-      if (ae)
+      if (err)
         {
-          log_error (_("assuan_inquire failed: %s\n"), assuan_strerror (ae));
-          err = map_assuan_err (ae);
+          log_error (_("assuan_inquire failed: %s\n"), gpg_strerror (err));
           goto leave;
         }
   
@@ -616,13 +661,13 @@ cmd_checkocsp (assuan_context_t ctx, char *line)
   if (!opt.allow_ocsp)
     err = gpg_error (GPG_ERR_NOT_SUPPORTED);
   else
-    err = ocsp_isvalid (ctrl, cert, NULL);
+    err = ocsp_isvalid (ctrl, cert, NULL, force_default_responder);
 
  leave:
   if (err)
     log_error (_("command %s failed: %s\n"), "CHECKOCSP", gpg_strerror (err));
   ksba_cert_release (cert);
-  return map_to_assuan_status (err);
+  return err;
 }
 
 
@@ -640,7 +685,6 @@ cmd_lookup (assuan_context_t ctx, char *line)
 {
   ctrl_t ctrl = assuan_get_pointer (ctx);
   gpg_error_t err = 0;
-  assuan_error_t ae;
   char *p;
   strlist_t sl, list = NULL;
   int truncated = 0, truncation_forced = 0;
@@ -742,15 +786,14 @@ cmd_lookup (assuan_context_t ctx, char *line)
           
           /* Send the data, flush the buffer and then send an END line
              as a certificate delimiter. */
-          ae = assuan_send_data (ctx, value, valuelen);      
-          if (!ae)
-            ae = assuan_send_data (ctx, NULL, 0);
-          if (!ae)
-            ae = assuan_write_line (ctx, "END");
-          if (ae) 
+          err = assuan_send_data (ctx, value, valuelen);      
+          if (!err)
+            err = assuan_send_data (ctx, NULL, 0);
+          if (!err)
+            err = assuan_write_line (ctx, "END");
+          if (err) 
             {
-              log_error (_("error sending data: %s\n"), assuan_strerror (ae));
-              err = map_assuan_err (ae);
+              log_error (_("error sending data: %s\n"), gpg_strerror (err));
               end_cert_fetch (fetch_context);
               goto leave;
             }
@@ -780,7 +823,7 @@ cmd_lookup (assuan_context_t ctx, char *line)
   if (err)
     log_error (_("command %s failed: %s\n"), "LOOKUP", gpg_strerror (err));
   free_strlist (list);
-  return map_to_assuan_status (err);
+  return err;
 }
 
 
@@ -807,7 +850,7 @@ cmd_loadcrl (assuan_context_t ctx, char *line)
   xfree (buf);
   if (err)
     log_error (_("command %s failed: %s\n"), "LOADCRL", gpg_strerror (err));
-  return map_to_assuan_status (err);
+  return err;
 }
 
 
@@ -825,13 +868,15 @@ cmd_listcrls (assuan_context_t ctx, char *line)
   gpg_error_t err;
   FILE *fp = assuan_get_data_fp (ctx);
 
+  (void)line;
+
   if (!fp)
     return PARM_ERROR (_("no data stream"));
 
   err = crl_cache_list (fp);
   if (err)
     log_error (_("command %s failed: %s\n"), "LISTCRLS", gpg_strerror (err));
-  return map_to_assuan_status (err);
+  return err;
 }
 
 
@@ -854,16 +899,16 @@ cmd_cachecert (assuan_context_t ctx, char *line)
   ctrl_t ctrl = assuan_get_pointer (ctx);
   gpg_error_t err;
   ksba_cert_t cert = NULL;
-  assuan_error_t ae;
   unsigned char *value = NULL;
   size_t valuelen; 
+
+  (void)line;
       
-  ae = assuan_inquire (ctrl->server_local->assuan_ctx, "TARGETCERT",
+  err = assuan_inquire (ctrl->server_local->assuan_ctx, "TARGETCERT",
                        &value, &valuelen, MAX_CERT_LENGTH);
-  if (ae)
+  if (err)
     {
-      log_error (_("assuan_inquire failed: %s\n"), assuan_strerror (ae));
-      err = map_assuan_err (ae);
+      log_error (_("assuan_inquire failed: %s\n"), gpg_strerror (err));
       goto leave;
     }
   
@@ -885,7 +930,7 @@ cmd_cachecert (assuan_context_t ctx, char *line)
   if (err)
     log_error (_("command %s failed: %s\n"), "CACHECERT", gpg_strerror (err));
   ksba_cert_release (cert);
-  return map_to_assuan_status (err);
+  return err;
 }
 
 
@@ -907,16 +952,16 @@ cmd_validate (assuan_context_t ctx, char *line)
   ctrl_t ctrl = assuan_get_pointer (ctx);
   gpg_error_t err;
   ksba_cert_t cert = NULL;
-  assuan_error_t ae;
   unsigned char *value = NULL;
   size_t valuelen; 
+
+  (void)line;
     
-  ae = assuan_inquire (ctrl->server_local->assuan_ctx, "TARGETCERT",
+  err = assuan_inquire (ctrl->server_local->assuan_ctx, "TARGETCERT",
                        &value, &valuelen, MAX_CERT_LENGTH);
-  if (ae)
+  if (err)
     {
-      log_error (_("assuan_inquire failed: %s\n"), assuan_strerror (ae));
-      err = map_assuan_err (ae);
+      log_error (_("assuan_inquire failed: %s\n"), gpg_strerror (err));
       goto leave;
     }
   
@@ -954,18 +999,18 @@ cmd_validate (assuan_context_t ctx, char *line)
   if (err)
     log_error (_("command %s failed: %s\n"), "VALIDATE", gpg_strerror (err));
   ksba_cert_release (cert);
-  return map_to_assuan_status (err);
+  return err;
 }
 
 
 
 /* Tell the assuan library about our commands. */
 static int
-register_commands (ASSUAN_CONTEXT ctx)
+register_commands (assuan_context_t ctx)
 {
   static struct {
     const char *name;
-    int (*handler)(ASSUAN_CONTEXT, char *line);
+    int (*handler)(assuan_context_t, char *line);
   } table[] = {
     { "ISVALID",    cmd_isvalid },
     { "CHECKCRL",   cmd_checkcrl },
@@ -996,6 +1041,8 @@ register_commands (ASSUAN_CONTEXT ctx)
 void
 start_command_handler (int fd)
 {
+  static const char hello[] = "Dirmngr " VERSION " at your service";
+  static char *hello_line;
   int rc;
   assuan_context_t ctx;
   ctrl_t ctrl;
@@ -1023,13 +1070,13 @@ start_command_handler (int fd)
     }
   else
     {
-      rc = assuan_init_connected_socket_server (&ctx, fd);
+      rc = assuan_init_socket_server_ext (&ctx, fd, 2);
     }
 
   if (rc)
     {
       log_error (_("failed to initialize the server: %s\n"),
-                 assuan_strerror(rc));
+                 gpg_strerror(rc));
       dirmngr_exit (2);
     }
 
@@ -1037,10 +1084,32 @@ start_command_handler (int fd)
   if (rc)
     {
       log_error (_("failed to the register commands with Assuan: %s\n"),
-                 assuan_strerror(rc));
+                 gpg_strerror(rc));
       dirmngr_exit (2);
     }
-  assuan_set_hello_line (ctx, "Dirmngr " VERSION " at your service");
+
+
+  if (!hello_line)
+    {
+      size_t n;
+      const char *cfgname;
+
+      cfgname = opt.config_filename? opt.config_filename : "[none]";
+
+      n = (30 + strlen (opt.homedir) + strlen (cfgname)
+           + strlen (hello) + 1);
+      hello_line = xmalloc (n+1);
+      snprintf (hello_line, n,
+                "Home: %s\n"
+                "Config: %s\n"
+                "%s",
+                opt.homedir,
+                cfgname,
+                hello);
+      hello_line[n] = 0;
+    }
+
+  assuan_set_hello_line (ctx, hello_line);
   assuan_register_option_handler (ctx, option_handler);
 
   ctrl->server_local->assuan_ctx = ctx;
@@ -1056,14 +1125,27 @@ start_command_handler (int fd)
         break;
       if (rc)
         {
-          log_info (_("Assuan accept problem: %s\n"), assuan_strerror (rc));
+          log_info (_("Assuan accept problem: %s\n"), gpg_strerror (rc));
           break;
         }
 
+#ifndef HAVE_W32_SYSTEM
+      if (opt.verbose)
+        {
+          pid_t apid;
+          uid_t auid;
+          gid_t agid;
+
+          if (!assuan_get_peercred (ctx, &apid, &auid, &agid))
+            log_info ("connection from process %ld (%ld:%ld)\n",
+                      (long)apid, (long)auid, (long)agid);
+        }
+#endif
+
       rc = assuan_process (ctx);
       if (rc)
         {
-          log_info (_("Assuan processing failed: %s\n"), assuan_strerror (rc));
+          log_info (_("Assuan processing failed: %s\n"), gpg_strerror (rc));
           continue;
         }
     }
@@ -1116,7 +1198,7 @@ dirmngr_status (ctrl_t ctrl, const char *keyword, ...)
             *p++ = *text++;
         }
       *p = 0;
-      err = map_assuan_err (assuan_write_status (ctx, keyword, buf));
+      err = assuan_write_status (ctx, keyword, buf);
     }
 
   va_end (arg_ptr);
