@@ -106,7 +106,7 @@ unknown_criticals (ksba_cert_t cert)
 static gpg_error_t
 check_cert_policy (ksba_cert_t cert)
 {
-  static char *allowed[] = {
+  static const char *allowed[] = {
     "2.289.9.9",
     NULL
   };
@@ -175,8 +175,22 @@ allowed_ca (ksba_cert_t cert, int *chainlen)
     return err;
   if (!flag)
     {
-      log_error (_("issuer certificate is not marked as a CA"));
-      return gpg_error (GPG_ERR_BAD_CA_CERT);
+      if (!is_trusted_cert (cert))
+        {
+          /* The German SigG Root CA's certificate does not flag
+             itself as a CA; thus we relax this requirement if we
+             trust a root CA.  I think this is reasonable.  Note, that
+             gpgsm implements a far stricter scheme here. */
+          if (chainlen)
+            *chainlen = 3; /* That is what the SigG implements. */
+          if (opt.verbose)
+            log_info (_("accepting root CA not marked as a CA"));
+        }
+      else
+        {
+          log_error (_("issuer certificate is not marked as a CA"));
+          return gpg_error (GPG_ERR_BAD_CA_CERT);
+        }
     }
   return 0;
 }
@@ -749,7 +763,11 @@ check_cert_sig (ksba_cert_t issuer_cert, ksba_cert_t cert)
       log_error (_("unknown hash algorithm `%s'\n"), algoid? algoid:"?");
       return gpg_error (GPG_ERR_GENERAL);
     }
-  s = gcry_md_algo_name (algo);
+  /* At some point in time a bug slipped into Libgcrypt returning
+     "ripemd160" as canonical name but only accepting "rmd160" in
+     pk_verify.  This will be fixed in libgcrypt 1.2.4 but we use a
+     workaround here. */
+  s = (algo == GCRY_MD_RMD160)? "rmd160": gcry_md_algo_name (algo);
   for (i=0; *s && i < sizeof algo_name - 1; s++, i++)
     algo_name[i] = tolower (*s);
   algo_name[i] = 0;   
